""" Handles requests and data untanglement towards retarded web service. """

from logging import getLogger
from .constants import (
    NO_CONTROLS,
    NO_ACCESS_RESPONSES
)
from .utils import (
    do_external_api_request,
    map_response_data,
)
from .models import (
    Account,
    AuthStatus,
)
from .exceptions import (
    NoAccessException,
    AuthenticationError,
)

_LOGGER = getLogger(__name__)


class Client:
    def __init__(self, user: Account | None = None):
        self.user = user

    def _api_request(self, params: dict, headers: dict | None = None):
        method = params.get('p')
        if method != 'l' and self.user:
            params = params | self.user.get_query_params()
        data = do_external_api_request(params, headers)
        if data in NO_ACCESS_RESPONSES:
            raise NoAccessException()
        _LOGGER.debug("Got response: %s", data)
        return data

    def set_user(self, user: Account):
        self.user = user

    def authenticate_user(self, username: str, password: str):
        auth_user = Account(**{"username": username, "password": password})
        params = {
            'p': 'l',
            'lang': 'no',
        } | auth_user.get_query_params()

        result = self._api_request(params)
        _LOGGER.debug("Got result: %s", result)
        user_dict = map_response_data(result, [
            'status',
            'country',
            None,
            'phone_prefix',
            'state',
            'uid',
        ])
        if user_dict.get('status') == AuthStatus.LOGIN_ERROR:
            raise AuthenticationError

        account_dict = {
            **user_dict
        } | auth_user.get_query_params()
        account = Account(**{str(k): str(v) for k, v in account_dict.items()})
        self.set_user(account)
        return account

    def get_settings(self):
        params = {
            'p': 'instillinger',
        }
        return self._api_request(params)

    def get_control(self, cid: int):
        params = {
            'p': 'hki',
            'kontroll_id': cid,
        }
        result = self._api_request(params)
        return map_response_data(result, [
            'id',            # 14241
            'county',        # Trøndelag
            'municipality',  # Malvik
            'type',          # Fartskontroll
            'timestamp',     # 29.05 - 20:47
            'description',   # Kontroll Olderdalen
            'lat',           # 63.4258007013951
            'lng',           # 10.6856604194473
            None,            # |
            None,            # |
            None,            # malvik.png
            None,            # trondelag.png
            'speed_limit',   # 90
            None,            # 1
            'last_seen',     # 20:47
            'confirmed',     # 0
            None,            # 2
            None,            # 1
        ])

    def get_controls(
        self,
        lat: float,
        lng: float,
        **kwargs,
    ):
        params = {
            'p': 'hk',
            **{
                'lat': lat,
                'lon': lng,
            },
            **kwargs,
        }

        result = self._api_request(params)
        if result == NO_CONTROLS:
            return []

        return map_response_data(result, [
            'id',           # 14239
            'county',       # Trøndelag
            'municipality', # Meråker
            'type',         # Toll/grense
            None,           # 20:02
            'description',  # Toll
            'lat',          # 63.3621679609569
            'lng',          # 11.9694197550416
            None,           # NOT_IN_USE
            None,           # meraaker.png
            None,           # YES
            None,           # meraaker.png
            'timestamp',    # 1685383334
            None,           # 0
            None,           # 20:04
            'last_seen',    # 1685383471
        ], multiple=True)

    def get_controls_in_radius(
        self,
        lat: float,
        lng: float,
        radius: int,
        speed: int = 100,
        **kwargs,
    ):
        return self.get_controls(lat, lng, p='gps_kontroller', vr=radius, speed=speed, **kwargs)

    def get_control_types(self):
        params = {
            'p': 'kontrolltyper',
        }
        result = self._api_request(params)

        data = map_response_data(result, [
            'slug',
            'name',
            'id',
            None,
        ], multiple=True)
        return [
            {
                # Remove ".png"
                key: val[:-4] if key == 'slug' else val
                for key, val in el.items()
            }
            for el in data
        ]

    def get_maps(self):
        params = {
            'p': 'hent_mine_kart',
        }
        result = self._api_request(params)
        data = map_response_data(result, [
            'id',
            None,
            'title',
            'country',
        ], multiple=True)
        return data

    def exchange_points(self):
        params = {
            'p': 'veksle',
        }
        result = self._api_request(params)
        data = map_response_data(result, [
            'status',
            'message',
        ])
        return data



import binascii
from datetime import datetime
from logging import getLogger
from json import JSONDecoder, JSONEncoder
import random
import string
import time
import base64
import re
from urllib.parse import urlencode
import requests
from Crypto.Cipher import AES

from .constants import (
    CRYPTO_K1,
    CRYPTO_K2,
    CLIENT_OS,
    CLIENT_VERSION_NUMBER,
    CLIENT_TIMEOUT,
    API_URL,
)

_LOGGER = getLogger(__name__)


def get_random_string(length: int, letters: str | None = None) -> str:
    if letters is None:
        letters = string.ascii_uppercase + string.digits
    return ''.join(random.choice(letters) for _ in range(length))


def generate_device_id():
    return get_random_string(16, letters=string.digits + 'abcdef')


def get_unix_timestamp():
    return int(time.time()) + 10


def hash_credentials(credentials: dict):
    creds = bytes(JSONEncoder().encode(credentials), 'utf-8')
    return base64.b64encode(creds).decode()


def unhash_credentials(credentials: str):
    creds = base64.b64decode(credentials).decode()
    return JSONDecoder().decode(creds)


def aes_encrypt(input_str: str):
    """
    Encrypts a string using AES encryption with given key and initialization vector.
    Returns base64-encoded result.
    """
    key = base64.b64decode(CRYPTO_K2)
    iv = base64.b64decode(CRYPTO_K1)
    cipher = AES.new(key, AES.MODE_CBC, iv)
    length = 16 - (len(input_str) % 16)
    input_b = bytes(input_str, 'utf-8') + bytes([length]) * length
    return base64.b64encode(cipher.encrypt(input_b)).decode()


def aes_decrypt(enc_base64: str):
    """
    Decrypts AES encrypted data using a given key and initialization vector.
    """
    enc_data = base64.b64decode(enc_base64)
    key = base64.b64decode(CRYPTO_K2)
    iv = base64.b64decode(CRYPTO_K1)
    decipher = AES.new(key, AES.MODE_CBC, iv)
    return decipher.decrypt(enc_data).decode()


def get_query_params(params: dict):
    """
    Generates a query dictionary with random and predefined values.
    Replaces special characters in the values with hyphens.
    """
    query = {
        'bac': get_random_string(10),
        'z': int(time.time()) + 10,
        'version': CLIENT_VERSION_NUMBER,
        'os': CLIENT_OS,
        **params,
        'tt': get_random_string(5),
    }
    for k, v in query.items():
        if k in ['bac', 'z', 'version', 'os', 'tt']:
            query[k] = re.sub(r"[|#\\\"]", "-", str(v))
    return query


def do_external_api_request(params: dict, headers: dict | None = None):
    if headers is None:
        headers = {}

    payload = get_query_params(params)
    _LOGGER.debug("Doing API request with params: %s", payload)
    url = f'{API_URL}/app.php?{aes_encrypt(urlencode(payload))}'
    r = requests.get(url, headers={
        'user-agent': f'PK_{CLIENT_VERSION_NUMBER}',
        **headers,
    }, timeout=CLIENT_TIMEOUT)
    r.raise_for_status()
    try:
        data = aes_decrypt(r.text)
    except binascii.Error:
        data = r.text

    return data.strip('\x00\x01\x02\x03\x04\x05\x06\x07\x08\x10\x0f').strip()


def map_response_data(
    data: str,
    map_keys: list[str | None],
    multiple=False
) -> list[dict[str, str]] | dict[str, str]:
    """Converts a cvs-like string into dictionaries."""

    def row_to_dict(row) -> dict[str, str]:
        r = dict(zip(map_keys, row.split('|')))  # pylint: disable=modified-iterating-list
        return {k: r[k] for k in r.keys() if isinstance(k, str)}

    if multiple:
        return list(map(row_to_dict, list(data.split('#'))))

    return row_to_dict(data)


def parse_time_format(text: str):
    today = datetime.today()
    try:
        return int(
            datetime.strptime(text, "%d.%m - %H:%M").replace(
                year=today.year,
            ).timestamp()
        )
    except ValueError:
        pass

    try:
        return int(
            datetime.strptime(text, "%H:%M").replace(
                year=today.year,
                month=today.month,
                day=today.day,
            ).timestamp()
        )
    except ValueError:
        pass
    return text



""" CLI tool """

from tabulate import tabulate
from rich import print as rprint
import click

from politikontroller_py import Client
from politikontroller_py.models import (
    PoliceControl,
)
from politikontroller_py.exceptions import AuthenticationError

TABULATE_DEFAULTS = {
    'tablefmt': 'rounded_grid',
}

client = Client()


@click.group()
@click.option('--username', '-u', envvar='POLITIKONTROLLER_USERNAME', type=int,
              required=True, help="Username (i.e. phone number)")
@click.password_option(envvar='POLITIKONTROLLER_PASSWORD', type=str, required=True,
                       confirmation_prompt=False, help='Password')
def cli(username: int, password: str):
    """
    Username and password can be defined using env vars:

    POLITIKONTROLLER_USERNAME
    POLITIKONTROLLER_PASSWORD
    """
    try:
        client.authenticate_user(47, username, password)
    except AuthenticationError as err:
        raise click.BadParameter("Authentication error",
                                 param_hint=["--username", "--password"]) from err


@cli.command('get-controls', short_help='get a list of all active controls.')
@click.option('--lat', required=True, help='Your position (latitude)')
@click.option('--lng', required=True, help='Your position (longitude)')
def get_controls(lat: float, lng: float):
    controls = client.get_controls(lat, lng)
    click.echo(tabulate(controls, **TABULATE_DEFAULTS))


@cli.command('get-controls-radius',
             short_help='get all active controls inside a radius.')
@click.option('--lat', type=float, required=True, help='Radius center (latitude)')
@click.option('--lng', type=float, required=True, help='Radius center (longitude)')
@click.option('--radius', type=int, required=True, metavar='km',
              help='Radius size in kilometers')
@click.option('--speed', type=int, required=False, metavar='km/h',
              help='Speed, unknown what this does')
def get_controls_in_radius(lat: float, lng: float, radius: int, speed: int):
    controls = client.get_controls_in_radius(lat, lng, radius, speed)
    click.echo(tabulate(controls, **TABULATE_DEFAULTS))


@cli.command('get-control', short_help='get details on a control.')
@click.argument('control_id', type=int, required=True)
def get_control(control_id: int):
    control = PoliceControl(**client.get_control(control_id))
    rprint(control)
